#include "TFT80x160.h"
#include "fonts\6x8_FH.h"
#include "fonts\7x12_FH.h"
#include "fonts\8x16_FH.h"
#include "fonts\30x38_FH.h"
/*-------------------------------------------------------------------------------------------------*/


typedef struct __attribute__((packed))
{
  WORD  bfType;
  DWORD bfSize;
  WORD  bfReserved1;
  WORD  bfReserved2;
  DWORD bfOffBits;

  DWORD biSize;
  DWORD  biWidth;
  DWORD  biHeight;
  WORD  biPlanes;
  WORD  biBitCount;
  DWORD biCompression;
  DWORD biSizeImage;
  DWORD  biXPelsPerMeter;
  DWORD  biYPelsPerMeter;
  DWORD biClrUsed;
  DWORD biClrImportant;
} BMP_HDR;

typedef struct __attribute__((packed))
{ BYTE F;			// 'F' first 2 bytes are always FH
  BYTE H;			// 'H' for FONT horizontal
  BYTE first;		// first defined code
  BYTE last;		// last  defined code
  BYTE width;		// width of character in dots
  BYTE height;		// height of character in dots
  BYTE bytewidth;	// width of character in bytes
  BYTE bytechar;	// number of bytes needed for each character
  BYTE dat[1];		// begin of character data
} FH_FONT;

FH_FONT *gFont      = Font_8x16;
BYTE     gZoomX     = 1;
BYTE     gZoomY     = 1;
WORD     gColorFG   = RGB565(255,255,  0);
WORD     gColorBG   = RGB565(  0,  0,255);
WORD     gAngle     = 0;
WORD     gTFTwidth  = TFTWIDTH;
WORD     gTFTheight = TFTHEIGHT;

/*-------------------------------------------------------------------------------------------------*/

void setTFTangle(WORD angle)
{
  if(angle == 0  ||  angle == 180)
   { gAngle     = angle;
     gTFTwidth  = TFTWIDTH;
     gTFTheight = TFTHEIGHT;
   }
  else if(angle == 90  ||  angle == 270)
   { gAngle     = angle;
     gTFTwidth  = TFTHEIGHT;
	 gTFTheight = TFTWIDTH;
   }
}

void setTFTfont(BYTE nr, BYTE zx, BYTE zy)
{
  FH_FONT *pf = NULL;

  switch(nr)
   { case 1: pf = Font_6x8;		break;
     case 2: pf = Font_7x12;	break;
     case 3: pf = Font_8x16;	break;
     case 4: pf = Font_30x38;	break;
   }

  if(pf != NULL  &&  pf->F == 'F'  &&  pf->H == 'H')
     gFont = pf;

  if(zx >= 1  &&  zx <= 4)
     gZoomX = zx;

  if(zy >= 1  &&  zy <= 4)
     gZoomY = zy;
}

void setTFTcolor(WORD fg, WORD bg)
{
  gColorFG = fg;
  gColorBG = bg;
}

/*-------------------------------------------------------------------------------------------------*/

void TFTinit(void)   // controller ST7735s
{
  write_cmd(0x11);	// delay_ms out
  delay_ms(120);


  // ST7735S Frame Rate Setting in normal mode: fosc/(((RTNA*2)+40)*(LINE+FPA+BPA+2))=80
  write_cmd(0xB1);	// fosc=850KHz
  write_dat(0x05);	// RTNA=5
  write_dat(0x3C);	// FPA=58
  write_dat(0x3C);	// BPA=58

  // ST7735S Frame Rate Setting in idle mode: fosc/(((RTNB*2)+40)*(LINE+FPB+BPB+2))=80
  write_cmd(0xB2);	// fosc=850KHz
  write_dat(0x05);	//RTNB=5
  write_dat(0x3C);	//FPB=58
  write_dat(0x3C);	//BPB=58

  // ST7735S Frame Rate Setting in parital mode (dot inverson): fosc/(((RTNC*2)+40)*(LINE+FPC+BPC+2))=80
  // ST7735S Frame Rate Setting in parital mode (column inverson): fosc/(((RTNC*2)+40)*(LINE+FPC+BPC+2))=80
  write_cmd(0xB3);	// fosc=850KHz
  write_dat(0x05);	// RTNC=5
  write_dat(0x3C);	// FPC=58
  write_dat(0x3C);	// BPC=58
  write_dat(0x05);	// RTND=5
  write_dat(0x3C);	// FPD=58
  write_dat(0x3C);	// BPD=58

  // ST7735S Display Inversion Control
  write_cmd(0xB4);	// Dot inversion: 20184019 modify from Sitronix initial code
  write_dat(0x07);	// 0xB4[2]=Inversion setting in normal mode
                  	// 0xB4[1]=Inversion setting in idle mode
                  	// 0xB4[0]=Inversion setting in partial mode

  // ST7735S Power on Sequence
  write_cmd(0xC0);	// power control 1
  write_dat(0xE9);	// {Par.3[0],Par.1[4:0]}=VRHP[5:0]=2->GVDD=4.6, Par.1[7:5]=AVDD [2:0]=6->AVDD=5.1
  write_dat(0x09);	// {Par.3[1],Par.2[4:0]}=VRHN[5:0]=2->GVCL=-4.6
  write_dat(0x04);	// Par.3[7:6]=MODE[1:0]=2x

  write_cmd(0xC1);	// power control 2
  write_dat(0xC5);	// Par.1 [1:0]=VGHBT[1:0]=0->VGH=2*AVDD+VGH25-0.5
                  	// Par.1 [3:2]=VGLSEL[1:0]=0->VGL=-7.5
                  	// Par.1 [7:6]=VGLSEL[1:0]=3->VGH25=2.4

  write_cmd(0xC2);	// power control 3
  write_dat(0x0D);	// (Sitronix initial) {Par.1 [7:6],Par.2[7:0]=DCA[9:0]=000000000'b->Booster set up cycle BCLK/1 BCLK/3 BCLK/1 BCLK/1 BCLK/1 in normal mode
  write_dat(0x00);	// Par.1[5:3]=SAPA[2:0]=001'b-> OP current is small in normal mode
                  	// Par.1[2:0]=APA[2:0]=001'b-> OP current is Large in normal mode

  write_cmd(0xC3);      //power control 4
  write_dat(0x8D);	// (Sitronix initial){Par.1 [7:6],Par.2[7:0]=DCB[9:0]=1001101010'b->Booster set up cycle BCLK/2 BCLK/1 BCLK/2 BCLK/2 BCLK/2 in idle mode
  write_dat(0x6A);	// Par.1[5:3]=SAPB[2:0]=001'b-> OP current is small in idle mode
                  	// Par.1[2:0]=APB[2:0]=011'b-> OP current is Medium in idle mode

  write_cmd(0xC4);	// power control 5
  write_dat(0x8D);	// (Sitronix initial){Par.1 [7:6],Par.2[7:0]=DCC[9:0]=1011101110'b->Booster set up cycle BCLK/2 BCLK/2 BCLK/2 BCLK/4 BCLK/2 in partial mode
  write_dat(0xEE);	// Par.1[5:3]=SAPC[2:0]=001'b-> OP current is small in partial mode

  // ST7735S VCOM
  write_cmd(0xC5);	// VCOM setting value
  write_dat(0x15);	// 0xC5[5:0]=010010'b -> VCOM=-0.875

  write_cmd(0x21);	// ST7735 Display Inversion on

  // ST7735 Gamma Sequence
  write_cmd(0xE0);	// Gamma setting value (Positive Polarity)
  write_dat(0x07);	// Par.1[5:0]=VRF0P[5:0]=000011'b (Variable Resistor VRHP)
  write_dat(0x0E);	// Par.2[5:0]=VOS0P[5:0]=011011'b (Variable Resistor VRLP)
  write_dat(0x08);	// Par.3[5:0]=PK0P[5:0]=010010'b  (Voltage of V3 grayscale)
  write_dat(0x07);	// Par.4[5:0]=PK1P[5:0]=010001'b  (Voltage of V4 grayscale)
  write_dat(0x10);	// Par.5[5:0]=PK2P[5:0]=111111'b  (Voltage of V12 grayscale)
  write_dat(0x07);	// Par.6[5:0]=PK3P[5:0]=111010'b  (Voltage of V20 grayscale)
  write_dat(0x02);	// Par.7[5:0]=PK4P[5:0]=111010'b  (Voltage of V28 grayscale)
  write_dat(0x07);	// Par.8[5:0]=PK5P[5:0]=110100'b  (Voltage of V36 grayscale)
  write_dat(0x09);	// Par.9[5:0]=PK6P[5:0]=101111'b  (Voltage of V44 grayscale)
  write_dat(0x0F);	// Par.10[5:0]=PK7P[5:0]=101011'b  (Voltage of V52 grayscale)
  write_dat(0x25);	// Par.11[5:0]=PK8P[5:0]=110000'b  (Voltage of V56 grayscale)
  write_dat(0x36);	// Par.12[5:0]=PK9P[5:0]=111010'b  (Voltage of V60 grayscale)
  write_dat(0x00);	// Par.13[5:0]=SELV0P[5:0]=000000'b  (Voltage of V0 grayscale)
  write_dat(0x08);	// Par.14[5:0]=SELV1P[5:0]=000001'b  (Voltage of V1 grayscale)
  write_dat(0x04);	// Par.15[5:0]=SELV62P[5:0]=000010'b  (Voltage of V62 grayscale)
  write_dat(0x10);	// Par.16[5:0]=SELV63P[5:0]=001001'b  (Voltage of V63 grayscale)

  write_cmd(0xE1);	// Gamma setting value (Negative Polarity)
  write_dat(0x0A);	// Par.1[5:0]=VRF0N[5:0]=000011'b (Variable Resistor VRHN)
  write_dat(0x0D);	// Par.2[5:0]=VOS0N[5:0]=011011'b (Variable Resistor VRLN)
  write_dat(0x08);	// Par.3[5:0]=PK0N[5:0]=010010'b  (Voltage of V3 grayscale)
  write_dat(0x07);	// Par.4[5:0]=PK1N[5:0]=010001'b  (Voltage of V4 grayscale)
  write_dat(0x0F);	// Par.5[5:0]=PK2N[5:0]=110010'b  (Voltage of V12 grayscale)
  write_dat(0x07);	// Par.6[5:0]=PK3N[5:0]=101111'b  (Voltage of V20 grayscale)
  write_dat(0x02);	// Par.7[5:0]=PK4N[5:0]=101010'b  (Voltage of V28 grayscale)
  write_dat(0x07);	// Par.8[5:0]=PK5N[5:0]=101111'b  (Voltage of V36 grayscale)
  write_dat(0x09);	// Par.9[5:0]=PK6N[5:0]=101110'b  (Voltage of V44 grayscale)
  write_dat(0x0F);	// Par.10[5:0]=PK7N[5:0]=101100'b  (Voltage of V52 grayscale)
  write_dat(0x25);	// Par.11[5:0]=PK8N[5:0]=111001'b  (Voltage of V56 grayscale)
  write_dat(0x35);	// Par.12[5:0]=PK9N[5:0]=111111'b  (Voltage of V60 grayscale)
  write_dat(0x00);	// Par.13[5:0]=SELV0N[5:0]=000000'b  (Voltage of V0 grayscale)
  write_dat(0x09);	// Par.14[5:0]=SELV1N[5:0]=000000'b  (Voltage of V1 grayscale)
  write_dat(0x04);	// Par.15[5:0]=SELV62N[5:0]=000001'b  (Voltage of V62 grayscale)
  write_dat(0x10);	// Par.16[5:0]=SELV63N[5:0]=001001'b  (Voltage of V63 grayscale)

  write_cmd(0xFC);	// Enable Gate power save mode
  write_dat(0xC0);	// 0xFC[7:6]=GCV_Enable[1:0]=10'b-> Gate Pump Clock Frequency disable
                  	// 0xFC[3:2]=CLK_Variable[1:0]=11'b-> Save Power Ability is Large

  write_cmd(0x3A);	// Color Mode
  write_dat(0x05);	// 65K Mode RGB565

  write_cmd(0x29);	// Display on

}

/*-------------------------------------------------------------------------------------------------*/

void TFTwindow(WORD xs, WORD ys, WORD xe, WORD ye)
{
  WORD tmp = 0;

  write_cmd(0x36);	// ST7735 Memory data access control
  switch(gAngle)
   { case 0:    write_dat(0xC8);  break;
     case 90:   write_dat(0x68);  break;
     case 180:  write_dat(0x08);  break;
     case 270:  write_dat(0xA8);  break;
   }

  if(gAngle == 90  ||  gAngle == 270)	// swap x <-> y
   { tmp=xs; xs=ys; ys=tmp;
     tmp=xe; xe=ye; ye=tmp;
	 tmp = 1;
   }

  xs += 0x1A;
  xe += 0x1A;
  write_cmd((tmp==0) ? 0x2A : 0x2B); // 0x2A = Column-  0x2B = Row address set
  write_dat(HI(xs));
  write_dat(LO(xs));
  write_dat(HI(xe));
  write_dat(LO(xe));

  ys += 0x01;
  ye += 0x01;
  write_cmd((tmp==0) ? 0x2B : 0x2A); // 0x2B = Row-  0x2A = Column address set
  write_dat(HI(ys));
  write_dat(LO(ys));
  write_dat(HI(ye));
  write_dat(LO(ye));

  write_cmd(0x2C);	// Memory write
}

/*-------------------------------------------------------------------------------------------------*/
BYTE buf[2];
void write_color(WORD color)
{
	buf[0] = HI(color);
	buf[1] = LO(color);
	write_dat_buf(buf, 2);
}

void TFTfill(WORD color, WORD xs, WORD ys, WORD xe, WORD ye)
{
  int anz = (xe-xs+1) * (ye-ys+1);


  TFTwindow(xs,ys, xe,ye);
  
  while(anz--)
	write_color(color);
  
}

/*-------------------------------------------------------------------------------------------------*/

void zoombyte(BYTE *pd, BYTE b, BYTE zoom)
{
  switch(zoom)
   { case 1:  pd[0] = b;
           break;

     case 2:  pd[0] = pd[1] = 0;
              if(b & 0x80) pd[0]  = 0b11000000;
              if(b & 0x40) pd[0] |= 0b00110000;
              if(b & 0x20) pd[0] |= 0b00001100;
              if(b & 0x10) pd[0] |= 0b00000011;
              if(b & 0x08) pd[1]  = 0b11000000;
              if(b & 0x04) pd[1] |= 0b00110000;
              if(b & 0x02) pd[1] |= 0b00001100;
              if(b & 0x01) pd[1] |= 0b00000011;
           break;

     case 3:  pd[0] = pd[1] = pd[2] = 0;
              if(b & 0x80)   pd[0]  = 0b11100000;
              if(b & 0x40)   pd[0] |= 0b00011100;
              if(b & 0x20) { pd[0] |= 0b00000011;
                             pd[1]  = 0b10000000; }
              if(b & 0x10)   pd[1] |= 0b01110000;
              if(b & 0x08)   pd[1] |= 0b00001110;
              if(b & 0x04) { pd[1] |= 0b00000001;
                             pd[2]  = 0b11000000; }
              if(b & 0x02)   pd[2] |= 0b00111000;
              if(b & 0x01)   pd[2] |= 0b00000111;
           break;

     case 4:  pd[0] = pd[1] = pd[2] = pd[3] = 0;
              if(b & 0x80) pd[0]  = 0b11110000;
              if(b & 0x40) pd[0] |= 0b00001111;
              if(b & 0x20) pd[1]  = 0b11110000;
              if(b & 0x10) pd[1] |= 0b00001111;
              if(b & 0x08) pd[2]  = 0b11110000;
              if(b & 0x04) pd[2] |= 0b00001111;
              if(b & 0x02) pd[3]  = 0b11110000;
              if(b & 0x01) pd[3] |= 0b00001111;
           break;
   }
}

void getlinedat(BYTE *pd, BYTE *ps, BYTE byteanz, BYTE zoom)
{
  while(byteanz--)
   { zoombyte(pd, *ps++, zoom);
     pd += zoom;
   }
}

void TFTtext(WORD xs, WORD ys, char *ptext)
{
  WORD  x,y, xe,ye, width,height, z;
  BYTE *pdat, *pline, b, msk, linedat[100];

  if(gFont != NULL)
   { width  = gFont->width  * gZoomX;
     height = gFont->height * gZoomY;
     ye     = ys + height - 1;

     while(*ptext)
      { if(*ptext >= gFont->first  &&  *ptext <= gFont->last)
         { pdat = gFont->dat + (int)gFont->bytechar * (*ptext - gFont->first);
           xe = xs + width  - 1;

           TFTwindow(xs,ys, xe,ye);

           for(y=0; y < gFont->height; y++)
            {
              getlinedat(linedat, pdat, gFont->bytewidth, gZoomX);
              pdat += gFont->bytewidth;

              for(z=0; z < gZoomY; z++)
               { pline = linedat;
                 msk   = 0x80;
                 for(x=0; x < width; x++)
                  { if(msk == 0x80)
                       b = *pline++;
                    write_color((b & msk) ? gColorFG : gColorBG);
                    msk = (msk == 0x01) ? 0x80 : msk >> 1;
                  }
               }
            }
         }
        ptext++;
        xs += width;
      }
   }
}

/*-------------------------------------------------------------------------------------------------*/

void TFTbitmap(WORD xs, WORD ys, const BYTE *pbitmap)
{
  BMP_HDR *pBMP = (BMP_HDR *)pbitmap;
  BYTE    *pdat, r,g,b, msk;
  WORD     w,x, width,height, byteperline, color[256];

  if(pBMP->bfType == 0x4D42 )	// always BM, no RLE compressed BMPs
   { if(((pBMP->biBitCount <= 8 || pBMP->biBitCount == 24) && pBMP->biCompression == 0)  ||  (pBMP->biBitCount == 16 && pBMP->biCompression == 3))
      { width  = pBMP->biWidth;
        height = pBMP->biHeight;
        byteperline = (width * pBMP->biBitCount + 31) / 32 * 4;
        if(pBMP->biBitCount <= 8)
         { if((w=pBMP->biClrUsed) == 0)
            { if(pBMP->biBitCount == 1)  w = 2;
              if(pBMP->biBitCount == 4)  w = 16;
              if(pBMP->biBitCount == 8)  w = 256;
            }
           pdat = (BYTE *)pbitmap + pBMP->bfOffBits - w*4;
           for(x=0; x < w; x++)
            { color[x] = RGB565(pdat[2], pdat[1], pdat[0]);
              pdat += 4;
            }
         }

        TFTwindow(xs,ys, xs+width-1, ys+height-1);
        do
         { pdat = (BYTE *)pbitmap + pBMP->bfOffBits + byteperline * --height;
           switch(pBMP->biBitCount)
            {
              case 1:  msk = 0x80;
                       for(x=0; x < width; x++)
                        { if(msk == 0x80)
                             b = *pdat++;
                          write_color((b & msk) ? color[1] : color[0]);
                          msk = (msk == 0x01) ? 0x80 : msk >> 1;
                        }
                     break;

              case 4:  for(x=0; x < width; x++)
                        { if(x & 1)
                             write_color(color[b & 0x0F]);
                          else
                           { b = *pdat++;
                             write_color(color[b >> 4]);
                           }
                        }
                     break;

              case 8:  for(x=0; x < width; x++)
                          write_color(color[*pdat++]);
                     break;

              case 16: for(x=0; x < width; x++)
                        { w  = *pdat++;
                          w += *pdat++ << 8;
                          write_color(w);
                        }
                     break;

              case 24: for(x=0; x < width; x++)
                        { b = *pdat++;
                          g = *pdat++;
                          r = *pdat++;
                          w = RGB565(r,g,b);
                          write_color(w);
                        }
                     break;
            }
         }
        while(height > 0);

      }
   }
}


